/// Render Module
pub mod animation;
///
/// all of the code pertaining to Rendering will be written here
///
/// this also includes all Bevy Components and Resources relating to rendering
pub mod shader;

use core::slice;

use bevy_ecs::prelude::*;
use glow::HasContext;

use self::animation::Animator;

/// OpenGL context Resource
#[derive(Resource)]
pub struct GL(pub std::sync::Arc<glow::Context>);

/// Default Vertex Array Object, Every 2d Square object with a texture would only need this
#[derive(Resource)]
pub struct DefaultVAO(pub glow::NativeVertexArray);

/// Default Shader, this will be much more likely to be different
///
/// for all systems The Default Shader is assumed to be the animation shader located at data/shaders/AnimationShader
#[derive(Resource)]
pub struct DefaultShader(pub shader::Shader);

// NOTE systems start here

// a square with texture coordinates
//
// in a line
const VERTICES: [f32; 20] = [
    //  X       Y       Z       xTex   yTex
    0.5, 0.5, 0.0, 1.0, 1.0, 0.5, -0.5, 0.0, 1.0, 0.0, -0.5, -0.5, 0.0, 0.0, 0.0, -0.5, 0.5, 0.0,
    0.0, 1.0,
];

// indices for the
const INDICES: [u32; 6] = [0, 1, 3, 1, 2, 3];

pub fn initialize_defaults(mut commands: Commands, gl: Res<GL>) {
    // create vertex array
    let vao = unsafe { Some(gl.0.create_vertex_array().unwrap()) };

    unsafe {
        // create buffers
        let vbo = Some(gl.0.create_buffer().unwrap());
        let ebo = Some(gl.0.create_buffer().unwrap());

        gl.0.bind_vertex_array(vao);

        gl.0.bind_buffer(glow::ARRAY_BUFFER, vbo);

        // pass buffers
        let data = core::slice::from_raw_parts(
            VERTICES.as_ptr() as *const u8,
            VERTICES.len() * core::mem::size_of::<f32>(),
        );

        gl.0.named_buffer_data_u8_slice(vbo.unwrap(), data, glow::STATIC_DRAW);

        gl.0.bind_buffer(glow::ELEMENT_ARRAY_BUFFER, ebo);

        let data = slice::from_raw_parts(
            INDICES.as_ptr() as *const u8,
            INDICES.len() * core::mem::size_of::<u32>(),
        );

        gl.0.named_buffer_data_u8_slice(ebo.unwrap(), data, glow::STATIC_DRAW);

        let size = core::mem::size_of::<f32>() as i32;

        // Vertex Info
        gl.0.enable_vertex_attrib_array(0);
        gl.0.vertex_attrib_pointer_f32(0, 3, glow::FLOAT, false, size * 5, 0);

        // Texture Coordinates
        gl.0.enable_vertex_attrib_array(1);
        gl.0.vertex_attrib_pointer_f32(1, 2, glow::FLOAT, false, size * 5, size * 3);

        // unbind buffers and arrays
        gl.0.bind_buffer(glow::ARRAY_BUFFER, None);

        gl.0.bind_vertex_array(None);
    }

    // insert Default Vao
    commands.insert_resource(DefaultVAO(vao.unwrap()));

    let shader = shader::Shader::new("AnimationShader", &gl.0);

    commands.insert_resource(DefaultShader(shader));
}

pub fn test_init(mut commands: Commands, gl: Res<GL>) {
    let mut animator = Animator::default();

    let spritesheet = animation::SpriteSheet::new("data/sprite1.png", &gl.0, 4, Vec::new());

    animator
        .spritesheets
        .insert(animation::AnimationType::Idle, spritesheet);

    commands.spawn(animator);
}

pub fn test_draw(
    mut query: Query<&mut Animator>,
    gl: Res<GL>,
    vao: Res<DefaultVAO>,
    shader: Res<DefaultShader>,
) {
    for mut animator in query.iter_mut() {
        animator.frame_time += 0.1;
        let frame_time = animator.frame_time.clone();
        let sheet = &animator.spritesheets.get(&animator.current).unwrap();

        shader.0.Use(&gl.0);
        shader
            .0
            .set_float(&gl.0, "frameCount", 1.0 / (sheet.cuts_x as f32));
        shader
            .0
            .set_float(&gl.0, "frameTime", frame_time.floor() + 1.0);
        shader.0.set_float(&gl.0, "scale", 1.0);

        // NOTE this is where you would put the position
        shader
            .0
            .set_vector2(&gl.0, "uniPos", cgmath::Vector2::new(0.0, 0.0));

        unsafe {
            gl.0.bind_vertex_array(Some(vao.0));
            gl.0.bind_texture(glow::TEXTURE_2D, Some(sheet.texture_id));

            gl.0.draw_elements(glow::TRIANGLES, 6, glow::UNSIGNED_INT, 0);
        }
    }
}
